#ifndef RecoBTag_CombinedSVAlg_h
#define RecoBTag_CombinedSVAlg_h

#include "DataFormats/GeometryVector/interface/GlobalVector.h"
#include "RecoBTag/CombinedSV/interface/DiscriminatorComputer.h"
#include "RecoBTag/CombinedSV/interface/TrackFilter.h"
#include "RecoBTag/CombinedSV/interface/VertexFilter.h"
#include "RecoBTag/CombinedSV/interface/TrackInfoBuilder.h"
#include "RecoBTag/CombinedSV/interface/PseudoVertexBuilder.h"
#include "RecoVertex/VertexPrimitives/interface/VertexReconstructor.h"
#include "RecoVertex/VertexPrimitives/interface/TransientVertex.h"
#include "DataFormats/Candidate/interface/ParticleKinematics.h"
#include "DataFormats/BTauReco/interface/CombinedSVInfo.h"
#include "DataFormats/BTauReco/interface/CombinedSVTrack.h"
#include "DataFormats/BTauReco/interface/CombinedSVVertex.h"
#include <vector>
#include <map>

// \class CombinedSVAlg
// 
// \short concrete implementation of the combined b-tag algorithm
// This is the concrete implementation of the combined b-tag alogrithm
// 
// \author Ulrich Kerzel, based on code by Christian Weiser
// \version first version on 19th April 2006

/**
 * current status:
 * -----------------------------------------------------------------------------
 *  - "first pass" completed, all code to compute the necessary
 *    variables is there
 *  - the code compiles and works on the level of
 *    few events (using 070pre4): the values of the variables look
 *    sensible at first glance, both JetTag
 *    and extended information are written to 
 *    the output file by the producer
 * 
 * current issues which need to be adressed:
 * -----------------------------------------------------------------------------
 * - confirm track cuts of tracks to be used for 
 *   building the "RecoVertex", "PseudoVertex", "NoVertex" and the tracks used
 *   for constructing the likelihood
 *
 * minor issues:
 * -----------------------------------------------------------------------------
 * - TransientVertex makes problems
 * - Create reflection dictionary for CombinedSVInfo
 *
 * - hitPattern() related functions are currently
 *   disabled, need to be put in once avaiable in a stable
 *   way in a pre-release
 *   (affects only the cut on #Si hits, not critical)  
 * 
 * Input:
 * ---------------------------------------------
 * - jet with track association
 * - primary vertex
 * - charm prior
 * - cuts for track and (secondary) vertex related quantities
 *   (if not specified, sensible defaults are given in the constructor)
 *
 * Output:
 * ----------------------------------------------
 * - "standard" JetTag object common to all b, tau tagging algorithms
 *
 * Program flow:
 * ---------------------------------------------
 *
 * setups (by user):
 * - track cuts, secondary vertex cuts
 * - charm prior
 * - primary vertex
 * - "calibration", i.e. PDFs used to determine likelihood value
 * 
 * inputs:
 * - jet with track association
 *
 * to run the tagging:
 * - do setups as above
 * - call "tag()" method  
 *   returns final discriminator value as "quick" output
 *
 * - retrieve output objects:
 *   - "general" information common to all taggers
 *   - specific information for combined b-tagger
 *
 * N.B. for "RecoVertex"
 * --------------------------------------------
 * In case secondary vertices have been found ("RecoVertex"),
 * the collection of secundary vertices returned by the 
 * inclusive vertex finder may hold more than one
 * vertex. As the inclusive finder is not optimised
 * to reconstruct the most probable B decay vertex
 * the vertices found may be a mixture of secondary
 * (= B decay) vertex, tertiary vertex, etc.
 * Hence, several track based quantities are computed w.r.t
 * all secondary vertices found (and not per
 * vertex) to get a better 'b-related' estimate.
 * In detail, these quantities are:
 * - mass of all tracks at all secundary vertices
 * - energy of all traks at all secondary vertics
 *   divided by energy of all tracks in jet
 * - pt, pL using above 
 * - mean rapidity of all tracks at all secondary vertices
 * - multiplicity
 * 
 * If the primary vertex is found again by the inclusive vertex finder,
 * it is removed from * the list by checking the number of tracks
 * used in the secondary vertex which are also used to build the 
 * primary vertex.  * The vertex is rejected if > 65% of the
 * tracks are shared.
 * 
 *
 * The main tag(Jet, primVertex) routine:
 * --------------------------------------------
 *
 * - determineVertexType():
 *   - try to find secondary vertices in jet: fitSecondaryVertex()
 *     - apply cuts to track selection: acceptTrack(track)
 *     - use provided inclusive secondary vertex finder from 
 *       KalmanTrimmedVertexFinder package (for now)
 *     - filter out primary vertex from list of secondary vertices
 *       and apply vertex cuts: acceptVertex(vertex)
 *     ==> if success, this jet is in class "RecoVertex"

 *   - if no secondary vertex is found try to combine tracks
 *     with large impact parameter to vertex.
 *     -> vtxBuilder_.buildVector( VertexType = PseudoVertex)
 *       => if success, this jet is in class "PseudoVertex"
 *
 *   - if neither RecoVertex nor PseudoVertex, this jet in in class "NoVertex"
 *     - combine tracks with vtxBuilder_.buildVector( VertexType=NoVertex )
 *       (no cuts applied on "vertex")
 *
 * - compute all variables needed for tagging: computeTaggingVariables()
 *   (depends on type of secondary vertex)
 *
 * - copy over variables used for likelihood in the three vertex
 *   categories into vector of pairs (varName, varValue)
 *   -> not all variables are used by each vertex category
 *   -> but this way easier to construct loop for final discriminator
 *
 * - loop over above vector of pairs, construct likelihoods.
 *   - determine separate values for b, c, udsg
 *     in bins of vertex category, jet pt, jet eta from PDFs 
 *     (PDFs are determined from simulation)
 *
 * - combine likelihoods to combined tagging variable
 *   - construct ratioC, ratioUDSG
 *      ratioC      =       bLikelihood
 *                    -----------------------------
 *                     bLikelihood + cLikelihood
 *
 *      ratioUDSG   =       bLikelihood
 *                    -----------------------------
 *                     bLikelihood + udsgLikelihood
 *
 *  - combine:
 *     discriminator = priorCharm*ratioC + priorUDSG*ratioUDSG
 *     -> this method returns this descriminator
 *
 *  - provide:
 *    -  JetTag object ("standard" information common to all b, tau taggig
 *                       algorithms)
 *            
 */

class MagneticField;

namespace reco {

  class Vertex;
  class CombinedSVAlg {

  public:
    typedef std::pair<reco::CombinedSVEnums::TaggingVariable,double> SingleVar;
    typedef std::vector<SingleVar>        SingleVarColl;

    CombinedSVAlg( const VertexReconstructor &,
                     const BTagCalibration & calibration );
    ~CombinedSVAlg();

    /** This is the main routine to be called by the user. It returns the 
     *  value of the discriminator as basic output information
     *  The magnetic field needs to be propagated down.
     */
    double tag( const TransientVertex & primVertex,
                const reco::ParticleKinematics & jet,
                const std::vector < reco::TransientTrack > & tracks,
                const MagneticField * magneticField ); 

    /** Fit for secondary vertices 
     *  -> public because in priciple of interest to others
     *  see futher comments in routine
     *    Input:
     *  -----------
     *  - collection of tracks to be fitted to 
     *    secondary vertices
     *  - primary vertex (as this one may be found again and
     *    needs to be filtered out from the list of obtained
     *    vertices)
     *
     *    Output:
     * -------------
     *  - collection of secondary vertices found in jet
     *
     */   
    std::vector<TransientVertex> fitSecondaryVertex(
       const TransientVertex & primVertex, const std::vector < reco::TransientTrack > & );

    /** Clean up after a specific jet has been tagged  i.e. clear maps, vectors,
     *  reset variable values
     *  specific to a given jet. In order to avoid setting cuts, calibrations (=PDFs),
     *  etc. one can use one instance to tag several jets however, 
     *  this requires to reset information specific to a given jet once the output is retrieved and 
     *  the next jet is going to be tagged.
     */
    void reset();       
    

    // define cuts and settings
    void setTrackFilter ( const TrackFilter & );
    void setTrackIpSignificanceMin2D ( double s );

    void setSecondaryVertexCuts( double massMax,
             double cutV0, double cutPV, int multMin );
    void setCharmCut ( double );
    // void setVertexFilter ( const VertexFilter & );

    void setCharmPrior(double charmPrior);

    /**
      * defines a track to be used in a secondary vertex.
      * return rapidity
      */
    void setUsedInSecondaryVertex ( const reco::TransientTrack &, const GlobalVector & );

    const std::vector<TransientVertex> & secVertices() const;
    
    void print();
    void setVerbosity(int verbose);

  private:
    /** determine if jet falls in category:
     *     RecoVertex   : secondary vertex found (at least one) which passes the cuts
     *     PseudoVertex : tracks with significant IP successfully combined to vertex
     *     NoVertex     : all other cases
     */
    btag::Vertices::VertexType determineVertexType();

    /**
     * compute all variables needed for tagging
     */
    void computeTaggingVariables();
    void computeFirstTrackAboveCharmMass();
    void computeFlightDistances ( const GlobalVector & );
    

    /** copy over all variables used to determine likelihood into new collection
     *  -> then able to construct likelihood in the same way for all three 
     * vertex types
     */
    SingleVarColl fillVarVector();  


    /** determine whether a given track passes pre-defined cuts
     *  Cave: cut on 2D impact parameter significance is
     *        different for the "NoVertex" case than for the
     *        other cases!
     * 
     *        for "NoVertex" : 0.0
     *        else           : trackIpSignificanceMin2DMin_
     *
     */
    bool acceptTrack( const reco::TransientTrack & track,
                     double ipSigni2DCut ); 

    /**
     * determine rapidity of a track w.r.t. a given direction
     * default value is 5.0
     */
    double  computeTrackRapidity( const reco::TransientTrack & track,
                                  const GlobalVector & pAll);


    double computeAngleKinGeom( const GlobalVector & pTracks );

    /**
     *  Fraction of tracks shared with the primary vertex.
     */
    double fractionSharedTracks ( const TransientVertex & v ) const;

    /**
     * determine value of the PDF
     *
     * Input:
     * --------
     * - taggingVar   : which variable should be analysed (e.g. vertex mass, etc)
     * - partonType   : for which parton (b,c,udsg)
     *
     * also uses:
     * - vertex category (RecoVertex, PseudoVertex, NoVertex)
     * - jet pt
     * - jet eta
     *
     * the routine fetches a histogram from the b-tagging calibration
     * (based on input information) and returns the appropiate bin
     * content as the value of the variable used in the likelihood
     * to determine the combined b-tag.
     */

    /*
    double getPDFValue(reco::CombinedSVEnums::TaggingVariable taggingVar,
                       double                                  variableValue,
                       combsv::Partons::PartonType      partonType);
                       */

    void fillPseudoVertexVars ( SingleVarColl & );
    void fillNoVertexVars ( SingleVarColl & );
    void fillRecoVertexVars ( SingleVarColl & );
    void fillRapidities ( SingleVarColl & );

    reco::CombinedSVTrack getTrackData ( const reco::TransientTrack & );

    vector < reco::TransientTrack > filterTracks 
      ( const std::vector<reco::TransientTrack> & );

    /**
     *  create TrackRefVector of all tracks in jet
     */
    vector < reco::TransientTrack > createTracks();

    void updateVertexInfo ( std::vector<TransientVertex>::const_iterator, 
                            const std::vector < reco::TransientTrack > & trks );

    std::vector < reco::TransientTrack > getSecondaryTracks() const;

    void saveKinematics ( const std::vector<reco::TransientTrack> & alltracks,
        const std::vector < reco::TransientTrack > & secondaries );


  private:
    VertexReconstructor * theVtxReconstructor;

    TransientVertex primaryVertex_; /** primary vertex of the event
                                  *  associated to this jet:
                                  *  - if several primay vertices found,
                                  *    use the one with highest pt**2
                                  *  - if none found, try to fit one
                                  *  - if none fitted, use beamspot 
                                  *     constraint
                                  */
    /**
     * secondary Vertex (or vertices)
     * - for "RecoVertex": fitted vertices using inclusive 
     *                     vertex finder
     * - for "PseudoVertex" : tracks attached to "vertex"
     * - for "NoVerex"      : same
     * 
     * N.B.: the PseudoVertex and NoVertex should _never_ be used
     *        to access quantities related to position, chi^2 and
     *        covariance matrix!
     *        They are constructed at the geometric origin (beamline)
     *        and have a large covariance matrix.
     *        -> see buildVertex() for details
     */   
    std::vector< TransientVertex > secondaryVertexColl_;  
    // std::map < TransientVertex, reco::CombinedSVVertex > evertices_;

    btag::Vertices::VertexType vertexType_;
    reco::ParticleKinematics jet_;
    std::vector < reco::TransientTrack > tracks_;
    std::map < reco::TransientTrack, reco::CombinedSVTrack > etracks_;

    // need magnetic field for 3D impact parameter
    // which needs to be propagated down
    const MagneticField * magneticField_;

    int    verbose_;

    TrackFilter trackFilter_; // class that defines what tracks to throw away
    TrackInfoBuilder trackInfoBuilder_;
    PseudoVertexBuilder vtxBuilder_;
    double trackIpSignificanceMin2DMin_;
    VertexFilter vertexFilter_;
    DiscriminatorComputer comp_;
    double vertexCharmCut_;               // in GeV
    
    // masses of needed particles (in GeV/c**2)
    double mPiPlus_;                         
    double mKPlus_;                         
    double mK0_;
    reco::CombinedSVInfo combinedBTagInfo_;

  }; // class CombinedSVAlg
} // namespace reco

#endif
