#ifndef _TRACKER_IMPACTPOINTEXTRAPOLATOR_H_
#define _TRACKER_IMPACTPOINTEXTRAPOLATOR_H_

#include "TrackingTools/GeomPropagators/interface/Propagator.h"
#include "DataFormats/GeometryVector/interface/GlobalPoint.h"
#include "DataFormats/GeometryCommonDetAlgo/interface/DeepCopyPointerByClone.h"

class FreeTrajectoryState;
class TrajectoryStateOnSurface;

/** Extrapolate to impact point with respect to vtx,
 *  i.e. point of closest approach to vtx in 3D.
 *  The surface of the returned TrajectoryStateOnSurface 
 *  is chosen centred on vtx;
 *  the axes of the local coordinate system (x_loc, y_loc, z_loc) are 
 *  z_loc // trajectory direction at impact point;
 *  x_loc normal to trajectory and along impact vector (impact point - vtx);
 *  y_loc forms a right-handed system with the other axes.
 */

class ImpactPointExtrapolator {

public:

  /// constructor with default geometrical propagator
  ImpactPointExtrapolator();
  /// constructor with user-supplied propagator
  ImpactPointExtrapolator(const Propagator& u);

  /// extrapolation with default (=geometrical) propagator
  TrajectoryStateOnSurface extrapolate(const FreeTrajectoryState& fts, 
				       const GlobalPoint& vtx) const;
  /// as above, but from TrajectoryStateOnSurface
  TrajectoryStateOnSurface extrapolate(const TrajectoryStateOnSurface tsos, 
				       const GlobalPoint& vtx) const;
  /// extrapolation with user-supplied propagator
  TrajectoryStateOnSurface extrapolate(const FreeTrajectoryState& fts, 
				       const GlobalPoint& vtx, 
				       const Propagator& u) const;
  /// as above, but from TrajectoryStateOnSurface
  TrajectoryStateOnSurface extrapolate(const TrajectoryStateOnSurface tsos, 
				       const GlobalPoint& vtx, 
				       const Propagator& u) const;

private:
  /// extrapolation of (multi) TSOS with (internal or user-supplied) propagator
  TrajectoryStateOnSurface extrapolateFullState(const TrajectoryStateOnSurface tsos, 
						const GlobalPoint& vtx, 
						const Propagator& u) const;
  /// extrapolation of (single) FTS with (internal or user-supplied) propagator
  TrajectoryStateOnSurface extrapolateSingleState(const FreeTrajectoryState& fts, 
						  const GlobalPoint& vtx, 
						  const Propagator& u) const;

private:
  DeepCopyPointerByClone<Propagator> thePropagator;
};

#endif
